@@ -228,6 +228,7 @@ public:
 private:
     ObjList *getSectHolder(const String& sect) const;
     ObjList *makeSectHolder(const String& sect);
+    bool loadFile(const char* file, String sect, unsigned int depth, bool warn);
     ObjList m_sections;
 };
 
@@ -487,7 +488,6 @@ private:
 class YATE_API MessageHandler : public String
 {
     friend class MessageDispatcher;
-    friend class MessageRelay;
     YNOCOPY(MessageHandler); // no automatic copies please
 public:
     /**
@@ -552,10 +552,16 @@ public:
     inline const NamedString* filter() const
 	{ return m_filter; }
 
+    /**
+     * Retrieve the Regexp filter (if installed) associated to this handler
+     */
+    inline const Regexp* filterRegexp() const
+	{ return m_filterRegexp; }
+
     /**
      * Set a filter for this handler
      * @param filter Pointer to the filter to install, will be owned and
-     *  destroyed by the handler
+     *  destroyed by the handler. The filter may be a NamedPointer carrying a Regexp
      */
     void setFilter(NamedString* filter);
 
@@ -579,14 +585,25 @@ protected:
      */
     void cleanup();
 
-private:
+    /**
+     * Internal use received function, do not call or override!
+     * @param msg The received message
+     * @return True if message was processed
+     */
     virtual bool receivedInternal(Message& msg);
-    void safeNow();
+
+    /**
+     * Internal use handler unlock, do not call!
+     */
+    void safeNowInternal();
+
+private:
     String m_trackName;
     unsigned m_priority;
     int m_unsafe;
     MessageDispatcher* m_dispatcher;
     NamedString* m_filter;
+    Regexp* m_filterRegexp;
     NamedCounter* m_counter;
 };
 
@@ -645,8 +662,14 @@ public:
     inline int id() const
 	{ return m_id; }
 
-private:
+    /**
+     * Internal use received function, do not call or override!
+     * @param msg The received message
+     * @return True if message was processed
+     */
     virtual bool receivedInternal(Message& msg);
+
+private:
     MessageReceiver* m_receiver;
     int m_id;
 };
@@ -773,6 +796,27 @@ public:
     inline void clear()
 	{ m_handlers.clear(); m_hookAppend = &m_hooks; m_hooks.clear(); }
 
+    /**
+     * Check if there is at least one message in the queue
+     * @return True if the queue holds at least one message
+     */
+    inline bool hasMessages() const
+	{ return m_messages.get() || m_messages.next(); }
+
+    /**
+     * Check if there is at least one handler installed
+     * @return True if at least one handler is installed
+     */
+    inline bool hasHandlers() const
+	{ return m_handlers.get() || m_handlers.next(); }
+
+    /**
+     * Check if there is at least one message hook installed
+     * @return True if at least one hook is installed
+     */
+    inline bool hasHooks() const
+	{ return m_hooks.get() || m_hooks.next(); }
+
     /**
      * Get the number of messages waiting in the queue
      * @return Count of messages in the queue
@@ -791,6 +835,51 @@ public:
      */
     unsigned int postHookCount();
 
+    /**
+     * Get the total number of enqueued messages
+     * @return Count of enqueued messages
+     */
+    u_int64_t enqueueCount() const
+	{ return m_enqueueCount; }
+
+    /**
+     * Get the total number of dequeued messages
+     * @return Count of dequeued messages
+     */
+    u_int64_t dequeueCount() const
+	{ return m_dequeueCount; }
+
+    /**
+     * Get the total number of dispatched messages
+     * @return Count of dispatched messages
+     */
+    u_int64_t dispatchCount() const
+	{ return m_dispatchCount; }
+
+    /**
+     * Get the queued messages high watermark
+     * @return Highest number of messages in queue
+     */
+    u_int64_t queuedMax() const
+	{ return m_queuedMax; }
+
+    /**
+     * Get the average dequeued message age in milliseconds or microseconds
+     * @param usec True to return microseconds instead of milliseconds
+     * @return Average age of dequeued messages
+     */
+    u_int64_t messageAge(bool usec = false) const
+	{ return usec ? m_msgAvgAge : ((m_msgAvgAge + 500) / 1000); }
+
+    /**
+     * Retrieve all statistics counters
+     * @param enqueued Returns count of enqueued messages
+     * @param dequeued Returns count of dequeued messages
+     * @param dispatched Returns count of all dispatched messages, including dequeued ones
+     * @param queueMax Returns queued high watermark
+     */
+    void getStats(u_int64_t& enqueued, u_int64_t& dequeued, u_int64_t& dispatched, u_int64_t& queueMax);
+
     /**
      * Install or remove a hook to catch messages after being dispatched
      * @param hook Pointer to a post-dispatching message hook
@@ -816,6 +905,11 @@ private:
     String m_trackParam;
     unsigned int m_changes;
     u_int64_t m_warnTime;
+    u_int64_t m_enqueueCount;
+    u_int64_t m_dequeueCount;
+    u_int64_t m_dispatchCount;
+    u_int64_t m_queuedMax;
+    u_int64_t m_msgAvgAge;
     int m_hookCount;
     bool m_hookHole;
 };
@@ -1349,6 +1443,13 @@ public:
      */
     static bool restart(unsigned int code, bool gracefull = false);
 
+    /**
+     * Check if the engine was started
+     * @return True if the engine has gone through the start phase
+     */
+    static bool started()
+	{ return s_started; }
+
     /**
      * Check if the engine is currently exiting
      * @return True if exiting, false in normal operation
@@ -1466,6 +1567,46 @@ public:
     inline unsigned int postHookCount()
 	{ return m_dispatcher.postHookCount(); }
 
+    /**
+     * Get the rate of dispatched messages per second
+     * @return Number of messages dispatched in the last second
+     */
+    inline unsigned int messageRate() const
+	{ return m_messageRate; }
+
+    /**
+     * Get the maximum rate of dispatched messages per second
+     * @return Maximum number of messages dispatched per second
+     */
+    inline unsigned int messageMaxRate() const
+	{ return m_maxMsgRate; }
+
+    /**
+     * Get the average dequeued message age in milliseconds or microseconds
+     * @param usec True to return microseconds instead of milliseconds
+     * @return Average age of dequeued messages
+     */
+    unsigned int messageAge(bool usec = false) const
+	{ return (unsigned int)m_dispatcher.messageAge(usec); }
+
+    /**
+     * Retrieve dispatcher's statistics counters
+     * @param enqueued Returns count of enqueued messages
+     * @param dequeued Returns count of dequeued messages
+     * @param dispatched Returns count of all dispatched messages, including dequeued ones
+     * @param queueMax Returns queued high watermark
+     */
+    inline void getStats(u_int64_t& enqueued, u_int64_t& dequeued, u_int64_t& dispatched, u_int64_t& queueMax)
+	{ m_dispatcher.getStats(enqueued,dequeued,dispatched,queueMax); }
+
+    /**
+     * Check if a plugin is currently loaded
+     * @param name Name of the plugin to check
+     * @return True if the named plugin is loaded
+     */
+    inline bool pluginLoaded(const String& name) const
+	{ return !!m_libs[name]; }
+
     /**
      * Loads the plugins from an extra plugins directory or just an extra plugin
      * @param relPath Path to the extra directory, relative to the main modules
@@ -1555,6 +1696,12 @@ private:
     void tryPluginFile(const String& name, const String& path, bool defload);
     ObjList m_libs;
     MessageDispatcher m_dispatcher;
+    uint64_t m_dispatchedLast;
+    unsigned int m_messageRate;
+    unsigned int m_maxMsgRate;
+    bool m_rateCongested;
+    bool m_queueCongested;
+    bool m_ageCongested;
     static Engine* s_self;
     static String s_node;
     static String s_shrpath;
@@ -1565,6 +1712,7 @@ private:
     static NamedList s_params;
     static int s_haltcode;
     static RunMode s_mode;
+    static bool s_started;
     static unsigned int s_congestion;
     static CallAccept s_accept;
     static const TokenDict s_callAccept[];
